import shutil

CINIT = '\033[{}m'
CEND = '\033[0m'
CBLOCK = '\033[{}m{}\033[0m'

CBOLD = '1'
CDIM = '2'
CITALIC = '3'
CURL = '4'
CSBLINK = '5'
CFBLINK = '6'
CSELECTED = '7'

CBLACK = '30'
CRED = '31'
CGREEN = '32'
CYELLOW = '33'
CBLUE = '34'
CVIOLET = '35'
CCYAN = '36'
CGREY = '37'
CWHITE = '38'

CBLACKBG = '40'
CREDBG = '41'
CGREENBG = '42'
CYELLOWBG = '43'
CBLUEBG = '44'
CVIOLETBG = '45'
CCYANBG = '46'
CGREYBG = '47'
CWHITEBG = '48'

CBLACKH = '90'
CREDH = '91'
CGREENH = '92'
CYELLOWH = '93'
CBLUEH = '94'
CVIOLETH = '95'
CCYANH = '96'
CGREYH = '97'
CWHITEH = '98'

CBLACKBGH = '100'
CREDBGH = '101'
CGREENBGH = '102'
CYELLOWBGH = '103'
CBLUEBGH = '104'
CVIOLETBGH = '105'
CCYANBGH = '106'
CGREYBGH = '107'
CWHITE = '108'

PGBLOCK = '\u2588'

HSOLID = '\u2500'
VSOLID = '\u2502'

TLSOLID = '\u250C'
TRSOLID = '\u2510'
TDSOLID = '\u252C'

BLSOLID = '\u2514'
BRSOLID = '\u2518'
BUSOLID = '\u2534'

CLSOLID = '\u251C'
CSOLID = '\u253C'
CRSOLID = '\u2524'

cstyles = {
    'bold': '1',
    'dim': '2',
    'italic': '3',
    'underline': '4',
    'black': CBLACK,
    'red': CRED,
    'green': CGREEN,
    'yellow': CYELLOW,
    'blue': CBLUE,
    'violet': CVIOLET,
    'cyan': CCYAN,
    'grey': CGREY,
    'white': CWHITE,
    'bgblack': CBLACKBG,
    'bgred': CREDBG,
    'bggreen': CGREENBG,
    'bgyellow': CYELLOWBG,
    'bgblue': CBLUEBG,
    'bgviolet': CVIOLETBG,
    'bgcyan': CCYANBG,
    'bggrey': CGREYBG,
    'bgwhite': CWHITE,
    'hblack': CBLACKH,
    'hred': CREDH,
    'hgreen': CGREENH,
    'hyellow': CYELLOWH,
    'hblue': CBLUEH,
    'hviolet': CVIOLETH,
    'hcyan': CCYANH,
    'hgrey': CGREYH,
    'hwhite': CWHITEH,
    'hbgblack': CBLACKBGH,
    'hbgred': CREDBGH,
    'hbggreen': CGREENBGH,
    'hbgyellow': CYELLOWBGH,
    'hbgblue': CBLUEBGH,
    'hbgviolet': CVIOLETBGH,
    'hbgcyan': CCYANBGH,
    'hbggrey': CGREYBGH,
    'hbgwhite': CWHITE,
}


def stext(text: str, style: str):
    splt = style.split(' ')
    ccode = ';'.join([cstyles.get(item, '39') for item in splt])
    return CBLOCK.format(ccode, text)


def sprint(text: str, style='', align='<', end='\n'):
    ste = stext(text, style=style)
    if align == '>':
        tml_columns = shutil.get_terminal_size().columns
        size = len(text)
        ste = ' ' * (tml_columns - size) + ste
    elif align == '^':
        tml_columns = shutil.get_terminal_size().columns
        ste = ste.center(tml_columns)
    print(stext, end=end)


def cmove(x, y):
    print('\033[{x};{y}HTESTE'.format(x=x, y=y))


class CTable:
    def __init__(self, columns: list[str]):
        self.columns = columns


def ctable(
        data: list,
        columns: list[str],
        style='',
        end_last=True,
        print_header=True
):
    tml_columns = shutil.get_terminal_size().columns
    tml_rest = tml_columns
    tcols = {}
    thline = f'{TLSOLID}'
    bhline = f'{CLSOLID}'
    iline = f'{CLSOLID}'
    bline = f'{BLSOLID}'
    htext = f'{VSOLID}'
    for idx, col in enumerate(columns):
        splt = col.split(':')
        name = splt[0]
        if len(splt) > 1:
            align = splt[1][-1] if splt[1][-1] in ['<', '^', '>'] else '<'
            size = int(splt[1][:-1] if splt[1][-1] in ['<', '^', '>'] else splt[1])
            tml_rest -= size
        else:
            align = '<'
            size = 0
        cl = {
            'name': name,
            'align': align,
            'size': size,
            'last': idx == len(columns) - 1
        }
        tcols[idx] = cl
    for idx, col in tcols.items():
        if col['size'] == 0:
            col['size'] = tml_rest - (len(tcols) + 1)
        size = col['size']
        last = col['last']
        name = col['name']
        align = col['align']
        thline += HSOLID * size + (TRSOLID if last else TDSOLID)
        bhline += HSOLID * size + (CRSOLID if last else CSOLID)
        iline += HSOLID * size + (CRSOLID if last else CSOLID)
        bline += HSOLID * size + (BRSOLID if last else BUSOLID)
        match align:
            case '<':
                htext += ' ' + name.ljust(size - 1) + VSOLID
            case '^':
                htext += name.center(size) + VSOLID
            case '>':
                htext += name.rjust(size - 1) + ' ' + VSOLID
    if print_header:
        sprint(thline, style=style)
        sprint(htext, style=style)
        sprint(bhline, style=style)
    for item in data:
        dline = VSOLID
        for idx, val in enumerate(item):
            col = tcols[idx]
            match col['align']:
                case '<':
                    dline += ' ' + str(val).ljust(col['size'] - 1) + VSOLID
                case '^':
                    dline += str(val).center(col['size']) + VSOLID
                case '>':
                    dline += str(val).rjust(col['size'] - 1) + ' ' + VSOLID
        sprint(dline, style=style)
    if end_last:
        sprint(bline, style=style)
    else:
        sprint(iline, style=style)


def cprogress(
        size: int,
        max=0.0,
        value=0.0,
        color='grey',
        loop=True,
        hold=False,
        before: str = '',
        after: str = ''
):
    fator = size / max
    maked_size = int(value * fator)
    rest_size = size - maked_size
    maked = stext(('\u2501' * maked_size), style=f'{color} bold')
    rest = stext(('\u2500' * rest_size), style=f'{color} dim')
    if loop:
        if rest_size > 0:
            print(f'{before} {maked}{rest} {after}', end='\r' if loop else '\n')
        elif hold:
            print(f'{before} {maked}{rest} {after}')
    else:
        print(f'{before} {maked}{rest} {after}')



from enum import Enum


class Option:
    class Separators(Enum):
        SPACE = 'S'
        EQUAL = '='

    def __init__(
            self,
            name: str,
            description: str,
            short: str = ''
    ):
        self.name = name
        self.description = description
        self.short = short
        self.var_name = ''
        self._var_type = bool
        self.value = None

    def __str__(self):
        return f'[{self.__class__.__name__}]{self.name}: {self.value}'

    def __repr__(self):
        return f'[{self.__class__.__name__}]{self.name}: {self.value}'

    @property
    def var_type(self):
        return self._var_type

    @var_type.setter
    def var_type(self, name):
        ret = None
        match name:
            case 'str': ret = str
            case 'int': ret = int
            case 'float': ret = float
            case _: ret = bool
        self._var_type = ret



import sys
from climy.command import Command
from climy.option import Option


class Application(Command):
    def __init__(
            self,
            name: str,
            description: str,
            title: str = '',
            version: str = '0.0.0',
            option_default_separator: Option.Separators = Option.Separators.EQUAL,
    ):
        super().__init__(name=name, description=description, title=title)
        self.version = version
        self.workdir: str = ''
        self.option_default_separator = option_default_separator

    def add_command(
            self,
            command: 'Command',
            *,
            name: str = ''
    ):
        command.set_app(self)
        super().add_command(command=command, name=name)

    def generate_help(self, text: str = ''):
        text += "\033[1m{title}\033[0m (version {version})".format(title=self.title, version=self.version)
        text = super().generate_help(text=f'{text}\r\n\r\n')
        return text

    def run(self):
        self.args = [*sys.argv]
        self.workdir = self.args.pop(0)
        super().run()



from climy.option import Option


class Handler:
    def __init__(self, command: 'Command'):
        self.command = command

    def run(self):
        ...


class Command:
    def __init__(
            self,
            name: str,
            description: str,
            title: str = '',
            handler=None
    ):
        self.commands: dict[str, Command] = {}
        self.options: dict[str, Option] = {}
        self.options_short: dict[str, str] = {}
        self.options_required: set[str] = set()
        self.name = name
        self.title = title
        self.description = description
        self.help_item_tabsize = 25
        self.args = ''
        self.vars = []
        self.handler = handler
        self.parent = None
        self.app = None

    def __str__(self):
        return f'{self.__class__.__name__}:{self.name}'

    def __repr__(self):
        return f'{self.__class__.__name__}:{self.name}'

    def add_command(
            self,
            command: 'Command',
            *,
            name: str = ''
    ):
        command.parent = self
        command.app = command.app or self.app
        self.commands[name or command.name] = command

    def add_option(
            self,
            name: str,
            description: str,
            *,
            short: str = '',
            var_type: str = 'bool',
            var_name: str = '',
            required: bool = False
    ):
        option = Option(name=name, description=description, short=short)
        option.var_name = var_name
        option.var_type = var_type
        option.required = required
        self.options[name] = option
        if short:
            self.options_short[short] = name
        if required:
            self.options_required.add(name)

    def run(self):
        if (not self.args and not self.handler) or (self.args and self.args[0] in ['-h', '--help', 'help']):
            text = self.generate_help()
            print(text)
            return
        if self.args:
            command = self.commands.get(self.args[0], None)
            if command:
                command.args = self.args[1:]
                command.run()
                return
        default_separator = self.app.option_default_separator
        args = iter(self.args)
        while True:
            try:
                arg = next(args)
            except Exception as e:
                break
            if default_separator == Option.Separators.EQUAL:
                pair = arg.split('=')
            else:
                pair = [arg]
            idx = 0
            if pair[0].startswith('--'):
                idx = 2
            elif pair[0].startswith('-'):
                idx = 1
            key = pair[0][idx:]
            name = self.options_short.get(key, key)
            option = self.options.get(name, None)
            if not option:
                self.vars = name
                continue
            if name in self.options_required:
                self.options_required.remove(name)
            if default_separator == Option.Separators.SPACE:
                option.value = True if option.var_type == 'bool' else next(args)
            else:
                if option.var_type == 'bool':
                    option.value = True
                else:
                    option.value = pair.pop(1) if len(pair) > 1 else None
        if self.options_required:
            print(f'\033[1;35mSome options are required: [{", ".join(self.options_required)}]')
            return
        if self.handler:
            if issubclass(self.handler, Handler):
                handler = self.handler(command=self)
                handler.run()
            else:
                self.handler(comm=self)

    def generate_help(self, text: str = ''):
        text += "\033[1mUsage:\033[0m\r\n  {name} <command> [options] [arguments]".format(name=self.get_name())
        if self.options:
            text += self.generate_help_options_session()
        if self.commands:
            text += "\r\n\r\n\033[1mCommands:\033[0m\r\n  {opt}".format(opt=self.generate_help_commands_session())
        return text

    @staticmethod
    def generate_help_session(name, content):
        text = "\r\n\r\n\033[1m{name}:\033[0m{content}".format(name=name, content=content)
        return text

    def generate_help_options_session(self):
        key = '-h, --help'
        desc = 'Print this help.'
        content = '\r\n  \033[36m{key: <25}\033[0m {desc}'.format(key=key, desc=desc)
        for option in self.options.values():
            key = f'-{option.short}, --{option.name}' if option.short else f'--{option.name}'
            if option.var_type != 'bool':
                var_name = option.var_name or option.name
                key += f'=[{var_name.upper()}]'
            item = '\r\n  \033[36m{key: <25}\033[0m {desc}'.format(key=key, desc=option.description)
            content += item
        text = self.generate_help_session('Options', content=content)
        return text

    def generate_help_commands_session(self):
        size = self.help_item_tabsize
        text = '\r\n  '.join([f'\033[36m{n.ljust(size)}\033[0m {c.description}' for n, c in self.commands.items()])
        return text

    def get_name(self):
        if self.parent:
            return f'{self.parent.get_name()} {self.name}'
        return self.name

    def set_app(self, app):
        self.app = app
        for command in self.commands.values():
            command.set_app(app)
